from PyQt5.QtGui import QIcon, QDesktopServices
from PyQt5.QtXml import QDomDocument
from qgis.core import QgsProject, QgsLayoutExporter, \
    QgsReadWriteContext, QgsLayoutItemMap, QgsPrintLayout, \
    QgsLayoutItemRegistry, QgsMessageLog
from PyQt5.QtWidgets import QAction, QApplication
import tempfile
import os.path
from PyQt5.QtCore import QUrl, Qt, QDate
from ..b64_conv import B64Convertor
import math
from ..db_store_layout_items import Loader as StoreLayoutLoader
from .. import paggf_version
import sip
import html


class DocumentGenerator(object):
    """Class load a layout"""
    def __init__(  # Attention aux valeurs par defaut doivent etre meme entre DocumentGenerator et LayoutValidator
        self, template_path, substitution_map=dict(), feature=None,
        load_feature_in_sub_map=False, output_pdf=False,
        output_png=False, output_filename=None, version=None
    ):
        self.template_path = template_path
        self.substitution_map = substitution_map
        self.feature = feature
        self.load_feature_in_sub_map = load_feature_in_sub_map
        self.layout_map = dict()
        self.output_pdf = output_pdf
        self.output_png = output_png
        self.output_filename = output_filename
        self.version = version

    def export_img_path(self, ext):
        """
        ext: pdf ou png
        """
        output_filename = 'export'

        if self.output_filename:
            output_filename = self.output_filename

        return os.path.join(
            tempfile.gettempdir(), "{}.{}".format(output_filename, ext))

    def export_png(self):
        img_path = self.export_img_path('png')
        export_settings = QgsLayoutExporter.ImageExportSettings()
        exporter = QgsLayoutExporter(self.layout)
        exporter.exportToImage(img_path, export_settings)
        return img_path

    def export_pdf(self):
        out_path = self.export_img_path('pdf')
        export_settings = QgsLayoutExporter.PdfExportSettings()
        exporter = QgsLayoutExporter(self.layout)
        exporter.exportToPdf(out_path, export_settings)
        return out_path

    def add_substitution(self, key, value):
        """
        Add an element to the substitution_map
        """
        self.substitution_map[key] = str(value)

    def add_substitutions(self, dico):
        """
        Add elements to the substitution_map
        """
        for key in dico:
            self.add_substitution(key, dico[key])

    def get_substitution_map(self):
        """
        Retourne un dict d tel que pour toute clef key, tout texte [key]
        sera remplace par d[key]

        Exemple si d = {'nom':'marc'}, alors le texte "Bonjour [nom]" sera
        remplace par "Bonjour marc".
        """
        return self.substitution_map

    def load_feature_in_sub_map_fct(self):
        for field in self.feature.fields():
            field_name = field.name()
            field_data = self.feature[field_name]

            if isinstance(field_data, QDate):
                self.substitution_map[field_name] = field_data.toString(
                    "dd.MM.yyyy")
            else:
                self.substitution_map[field_name] = str(field_data)

    def generate_layout(self):
        if self.load_feature_in_sub_map:
            self.load_feature_in_sub_map_fct()

        template = QDomDocument()
        with open(self.template_path, encoding="utf-8") as template_file:
            data = template_file.read()

            for k in self.substitution_map:
                data = data.replace(
                    '[' + k + ']',
                    html.escape(self.substitution_map[k]))
            template.setContent(data)
        template_file.close()

        self.layout = QgsPrintLayout(QgsProject.instance())

        rw_context = QgsReadWriteContext()
        layout_items = self.layout.loadFromTemplate(
            template, rw_context)

        for cm in self.layout.items():
            if(cm.type() == QgsLayoutItemRegistry.LayoutMap):
                self.layout_map[cm.displayName()] = cm

    def get_callback_to_store_into_feature(
            self, layer, feature, field_pdf_name, field_png_name):
        def callback(pv_pdf_path, pv_png_path):
            layer.startEditing()

            if field_pdf_name:
                b64data_pdf = B64Convertor.base64_from_file(pv_pdf_path)
                ok = layer.changeAttributeValue(
                    feature.id(),
                    layer.fields().indexFromName(field_pdf_name), b64data_pdf)
                if not ok:
                    QgsMessageLog.logMessage(
                        'Error when saving field {}'.format(field_pdf_name),
                        'paggf')

            if field_png_name:
                b64data_png = B64Convertor.base64_from_file(pv_png_path)
                ok = layer.changeAttributeValue(
                    feature.id(),
                    layer.fields().indexFromName(field_png_name), b64data_png)
                if not ok:
                    QgsMessageLog.logMessage(
                        'Error when saving field {}'.format(field_png_name),
                        'paggf')

            # TODO ameliorer cette partie
            # ici chargement du xml pour la carte à partir de la db
            if self.version:
                if self.version in [paggf_version.BM, paggf_version.CID]:
                    db_layout_storage = StoreLayoutLoader.get(self.version)
                    db_layout_storage.store(feature, layer, self.layout)

        return callback

    def exec_(self):
        self.generate_layout()


class LayoutValidator(DocumentGenerator):
    """Class to open a layout then validate its export"""

    def __init__(
        self, feature, template_path, iface, substitution_map=dict(),
        zoom_to_feature=False, load_feature_in_sub_map=False, callback=None,
        zoom_to_feature_scale=1.1, scale_step=None, output_pdf=False,
        output_png=False, layers_config=dict(), output_filename=None,
    ):
        super().__init__(
            template_path, substitution_map, feature,
            load_feature_in_sub_map, output_pdf, output_png, output_filename)
        self.iface = iface
        self.zoom_to_feature = zoom_to_feature
        self.zoom_to_feature_scale = zoom_to_feature_scale
        self.callback = callback
        self.output_png = output_png
        self.output_pdf = output_pdf
        self.scale_step = scale_step
        self.output_filename = output_filename
        self.version = None
        self.layout_map = dict()

        # a dict (indexed form name of LayoutItemMap) containing
        # list of layers + path to a style file
        # ie [(l, 'xx.qml'), (l2, None)]
        self.layers_config = layers_config

    def exec_(self):
        """
        Affichage du layout puis quand valider creation de l'image
        """
        self.generate_layout()
        self.layout.refresh()
        self.layout_iface = self.iface.openLayoutDesigner(self.layout)

        if self.zoom_to_feature:
            layout_map_item = self.layout.referenceMap()

            if self.feature.geometry().isNull():
                self.layout_iface.messageBar().pushWarning(
                    'Impossibilité de centrer la carte sur l\'élément',
                    'la géométrie de l\'élément est nulle'
                )
            elif layout_map_item:
                bb = self.feature.geometry().boundingBox()
                bb.scale(self.zoom_to_feature_scale)
                layout_map_item.zoomToExtent(bb)

                if self.scale_step:
                    scale = layout_map_item.scale()
                    rounded_scale = math.ceil(scale / self.scale_step)\
                        * self.scale_step

                    layout_map_item.setScale(rounded_scale)

        for layout_map_name in self.layers_config:
            layout_map_raw = self.layout.itemById(layout_map_name)
            layout_map = sip.cast(layout_map_raw, QgsLayoutItemMap)
            layers_config = self.layers_config[layout_map_name]

            if layers_config:
                layout_map.setLayers([l[0] for l in layers_config])

                layer_style_overrides = layout_map.layerStyleOverrides()

                for conf in layers_config:
                    if len(conf) == 3:
                        layer, style_file_path, substitution_map = conf
                    else:
                        layer, style_file_path = conf
                        substitution_map = None

                    if layer is not None and style_file_path is not None:
                        with open(style_file_path, encoding="utf-8") \
                                as style_file:
                            style = style_file.read()

                            if substitution_map:
                                for key in substitution_map:

                                    style = style.replace(
                                        '[' + key + ']',
                                        substitution_map[key](self.feature)
                                    )

                            layer_style_overrides[layer.id()] = style

                        style_file.close()

                layout_map.setLayerStyleOverrides(layer_style_overrides)
                layout_map.setKeepLayerSet(True)  # force to use the layerSet
                layout_map.setKeepLayerStyles(True)  # force to use the style

        self.check_layout_window = self.layout_iface.parent()
        self.check_layout_window.destroyed.connect(self.close_callback)

        toolbar = self.check_layout_window.addToolBar('sif-td')
        icon = QIcon(
            os.path.join(
                os.path.dirname(__file__), "..", "icons", "icon_validate.png"))
        action = QAction(
            icon, 'Valider le layout', self.check_layout_window)
        action.triggered.connect(self.exec_callback)
        action.setEnabled(True)
        toolbar.addAction(action)

    def close_callback(self, event):
        # to execute when the dialog is closed
        pass

    def exec_callback(self):
        """
        Callback quand user clique sur bouton
        """
        QApplication.setOverrideCursor(Qt.WaitCursor)

        output_png_path = None
        if self.output_png:
            output_png_path = self.export_png()

        output_pdf_path = None
        if self.output_pdf:
            output_pdf_path = self.export_pdf()

        QApplication.restoreOverrideCursor()

        self.check_layout_window.close()

        if self.callback:
            self.callback(output_pdf_path, output_png_path)
        else:
            if output_pdf_path:
                # Default behavior
                QDesktopServices.openUrl(
                    QUrl.fromLocalFile(output_pdf_path))
            elif output_png_path:
                # Default behavior
                QDesktopServices.openUrl(
                    QUrl.fromLocalFile(output_png_path))
